/**
 * Copyright 2005 Jason LaDere
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Originally based on Project64 code.
 *
 */

/*
 * Main.java
 *
 * Created on December 3, 2005, 8:57 AM
 *
 * To change this template, choose Tools | Options and locate the template under
 * the Source Creation and Management node. Right-click the template and choose
 * Open. You can then make changes to the template in the Source Editor.
 */

/*
 * Project 64 - A Nintendo 64 emulator.
 *
 * (c) Copyright 2001 zilmar (zilmar@emulation64.com) and
 * Jabo (jabo@emulation64.com).
 *
 * pj64 homepage: www.pj64.net
 *
 * Permission to use, copy, modify and distribute Project64 in both binary and
 * source form, for non-commercial purposes, is hereby granted without fee,
 * providing that this license information and copyright notice appear with
 * all copies and any derived work.
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event shall the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Project64 is freeware for PERSONAL USE only. Commercial users should
 * seek permission of the copyright holders first. Commercial use includes
 * charging money for Project64 or software derived from Project64.
 *
 * The copyright holders request that bug fixes and improvements to the code
 * should be forwarded to them so if they want them.
 *
 */

package jario64;

import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.util.Properties;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.UIManager;
import jario64.Cop0.TlbException;
import jario64.Memory.MemoryException;
import jario64.savechips.Eeprom;
import jario64.savechips.FlashRam;
import jario64.savechips.Mempak;
import jario64.savechips.Sram;
import jario64.debug.Debug;
import java.io.InputStream;
import java.net.URL;
import resources.ResourceAnchor;

/**
 *
 * @author Jason
 */
public class Main {
    
    private static final boolean DEBUG = false;
    
    public static final boolean Default_ShowUnhandledMemory = true; //false;
    public static final boolean Default_ShowPifRamErrors = true; //false;
    public static final boolean Default_ShowTLBMisses = false;
    public static final boolean Default_UseTlb = true;
    public static final boolean Default_CacheInstructions = false;
    public static final boolean Default_UseAudio = true;
    public static final boolean Default_FrameLimit = true;
    
    public static final String AUTO_SAVE_DIR = "./Save/";
    
    public static final int R4300I_OPCODE = 0;
    public static final int R4300I_SPECIAL = 1;
    public static final int R4300I_REGIMM = 2;
    public static final int R4300I_COP0 = 3;
    public static final int R4300I_COP0_FUNCTION = 4;
    public static final int R4300I_COP1 = 5;
    public static final int R4300I_COP1_BC = 6;
    public static final int R4300I_COP1_S = 7;
    public static final int R4300I_COP1_D = 8;
    public static final int R4300I_COP1_W = 9;
    public static final int R4300I_COP1_L = 10;
    
    private static final int Default_CountPerOp = 2;
    
    public static Sram sram;
    public static Mempak mempak;
    public static FlashRam flashRam;
    public static Eeprom eeprom;
    
    // used by Registers, Memory, Dma
    public static Registers regs;
    public static Memory mem;
    public static Pif pif;
    public static Dma dma;
    
    public static Rsp rsp;
    public static Video video;
    public static Audio audio;
    
    public static Cpu cpu;
    
    // used by Dma
    public static boolean ShowUnhandledMemory = Default_ShowUnhandledMemory;
    
    // used by InterpreterOps
    public static boolean ShowTLBMisses = Default_ShowTLBMisses;
    
    // used by InterpreterOps, Cpu
    public static boolean UseTlb = Default_UseTlb;
    public static boolean CacheInstructions = Default_CacheInstructions;
    
    // used by Pif, Eeprom
    public static boolean ShowPifRamErrors = Default_ShowPifRamErrors;
    
    public static boolean UseAudio = Default_UseAudio;
    public static boolean FrameLimit = Default_FrameLimit;
    
    private static long cpuStartTime;
    private static JFrame hMainWindow;
    private static JPanel panel;
    
    /** Creates a new instance of Main */
    public Main() {
    }

    // called by InterpreterOps, Cpu
    public static void exit(String msg) {
        CloseCpu();
        long endTime = System.currentTimeMillis();
        float totalTime = endTime - cpuStartTime;
        long millisInCpu = (endTime - cpuStartTime) - video.millisInGfx - rsp.millisInDlist - rsp.millisInAlist - mem.millisInAudio;
//        System.out.println("RSP %: " + ((rsp.millisInRsp/totalTime)*100f));
        System.out.println("DLIST %: " + ((rsp.millisInDlist/totalTime)*100f));
        System.out.println("ALIST %: " + ((rsp.millisInAlist/totalTime)*100f));
        System.out.println("GFX   %: " + ((video.millisInGfx/totalTime)*100f));
        System.out.println("SND   %: " + ((mem.millisInAudio/totalTime)*100f));
        System.out.println("CPU   %: " + ((millisInCpu/totalTime)*100f));
        System.out.printf("%s\n", msg);
        Debug.close();
        System.exit(0);
    }
    
    private static void CloseCpu() {
        if (eeprom != null)
            eeprom.close();
        if (mempak != null)
            mempak.close();
        if (sram != null)
            sram.close();
        if (video.gfxPlugin != null)
            video.gfxPlugin.romClosed();
	if (audio.audioPlugin != null)
            audio.audioPlugin.romClosed();
        if (pif.controllerPlugin != null)
            pif.controllerPlugin.romClosed();
        rsp.shutdown();
    }
    
    private static Runnable[][] buildCpuOps(Cpu cpu, Cop0 cop0, Cop1 cop1) {
        Runnable[][] ops = new Runnable[11][0];
        
        ops[R4300I_OPCODE] = new Runnable[64];
        ops[R4300I_SPECIAL] = new Runnable[64];
        ops[R4300I_REGIMM] = new Runnable[32];
        ops[R4300I_COP0] = new Runnable[32];
        ops[R4300I_COP0_FUNCTION] = new Runnable[64];
        ops[R4300I_COP1] = new Runnable[32];
        ops[R4300I_COP1_BC] = new Runnable[32];
        ops[R4300I_COP1_S] = new Runnable[64];
        ops[R4300I_COP1_D] = new Runnable[64];
        ops[R4300I_COP1_W] = new Runnable[64];
        ops[R4300I_COP1_L] = new Runnable[64];
        
        ops[R4300I_OPCODE][ 0] = cpu.R4300i_opcode_SPECIAL;
        ops[R4300I_OPCODE][ 1] = cpu.R4300i_opcode_REGIMM;
        ops[R4300I_OPCODE][ 2] = cpu.r4300i_J;
        ops[R4300I_OPCODE][ 3] = cpu.r4300i_JAL;
        ops[R4300I_OPCODE][ 4] = cpu.r4300i_BEQ;
        ops[R4300I_OPCODE][ 5] = cpu.r4300i_BNE;
        ops[R4300I_OPCODE][ 6] = cpu.r4300i_BLEZ;
        ops[R4300I_OPCODE][ 7] = cpu.r4300i_BGTZ;
        ops[R4300I_OPCODE][ 8] = cpu.r4300i_ADDI;
        ops[R4300I_OPCODE][ 9] = cpu.r4300i_ADDIU;
        ops[R4300I_OPCODE][10] = cpu.r4300i_SLTI;
        ops[R4300I_OPCODE][11] = cpu.r4300i_SLTIU;
        ops[R4300I_OPCODE][12] = cpu.r4300i_ANDI;
        ops[R4300I_OPCODE][13] = cpu.r4300i_ORI;
        ops[R4300I_OPCODE][14] = cpu.r4300i_XORI;
        ops[R4300I_OPCODE][15] = cpu.r4300i_LUI;
        ops[R4300I_OPCODE][16] = cpu.R4300i_opcode_COP0;
        ops[R4300I_OPCODE][17] = cpu.R4300i_opcode_COP1;
        ops[R4300I_OPCODE][18] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_OPCODE][19] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_OPCODE][20] = cpu.r4300i_BEQL;
        ops[R4300I_OPCODE][21] = cpu.r4300i_BNEL;
        ops[R4300I_OPCODE][22] = cpu.r4300i_BLEZL;
        ops[R4300I_OPCODE][23] = cpu.r4300i_BGTZL;
        ops[R4300I_OPCODE][24] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_OPCODE][25] = cpu.r4300i_DADDIU;
        ops[R4300I_OPCODE][26] = cpu.r4300i_LDL;
        ops[R4300I_OPCODE][27] = cpu.r4300i_LDR;
        ops[R4300I_OPCODE][28] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_OPCODE][29] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_OPCODE][30] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_OPCODE][31] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_OPCODE][32] = cpu.r4300i_LB;
        ops[R4300I_OPCODE][33] = cpu.r4300i_LH;
        ops[R4300I_OPCODE][34] = cpu.r4300i_LWL;
        ops[R4300I_OPCODE][35] = cpu.r4300i_LW;
        ops[R4300I_OPCODE][36] = cpu.r4300i_LBU;
        ops[R4300I_OPCODE][37] = cpu.r4300i_LHU;
        ops[R4300I_OPCODE][38] = cpu.r4300i_LWR;
        ops[R4300I_OPCODE][39] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_LWU;
        ops[R4300I_OPCODE][40] = cpu.r4300i_SB;
        ops[R4300I_OPCODE][41] = cpu.r4300i_SH;
        ops[R4300I_OPCODE][42] = cpu.r4300i_SWL;
        ops[R4300I_OPCODE][43] = cpu.r4300i_SW;
        ops[R4300I_OPCODE][44] = cpu.r4300i_SDL;
        ops[R4300I_OPCODE][45] = cpu.r4300i_SDR;
        ops[R4300I_OPCODE][46] = cpu.r4300i_SWR;
        ops[R4300I_OPCODE][47] = cpu.r4300i_CACHE;
        ops[R4300I_OPCODE][48] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_LL;
        ops[R4300I_OPCODE][49] = cpu.r4300i_LWC1;
        ops[R4300I_OPCODE][50] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_OPCODE][51] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_OPCODE][52] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_OPCODE][53] = cpu.r4300i_LDC1;
        ops[R4300I_OPCODE][54] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_OPCODE][55] = cpu.r4300i_LD;
        ops[R4300I_OPCODE][56] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_SC;
        ops[R4300I_OPCODE][57] = cpu.r4300i_SWC1;
        ops[R4300I_OPCODE][58] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_OPCODE][59] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_OPCODE][60] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_OPCODE][61] = cpu.r4300i_SDC1;
        ops[R4300I_OPCODE][62] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_OPCODE][63] = cpu.r4300i_SD;
        
        ops[R4300I_SPECIAL][ 0] = cpu.r4300i_SPECIAL_SLL; // NOP?
        ops[R4300I_SPECIAL][ 1] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_SPECIAL][ 2] = cpu.r4300i_SPECIAL_SRL;
        ops[R4300I_SPECIAL][ 3] = cpu.r4300i_SPECIAL_SRA;
        ops[R4300I_SPECIAL][ 4] = cpu.r4300i_SPECIAL_SLLV;
        ops[R4300I_SPECIAL][ 5] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_SPECIAL][ 6] = cpu.r4300i_SPECIAL_SRLV;
        ops[R4300I_SPECIAL][ 7] = cpu.r4300i_SPECIAL_SRAV;
        ops[R4300I_SPECIAL][ 8] = cpu.r4300i_SPECIAL_JR;
        ops[R4300I_SPECIAL][ 9] = cpu.r4300i_SPECIAL_JALR;
        ops[R4300I_SPECIAL][10] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_SPECIAL][11] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_SPECIAL][12] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_SPECIAL_SYSCALL;
        ops[R4300I_SPECIAL][13] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_SPECIAL_BREAK; // ??
        ops[R4300I_SPECIAL][14] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_SPECIAL][15] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_SPECIAL_SYNC;
        ops[R4300I_SPECIAL][16] = cpu.r4300i_SPECIAL_MFHI;
        ops[R4300I_SPECIAL][17] = cpu.r4300i_SPECIAL_MTHI;
        ops[R4300I_SPECIAL][18] = cpu.r4300i_SPECIAL_MFLO;
        ops[R4300I_SPECIAL][19] = cpu.r4300i_SPECIAL_MTLO;
        ops[R4300I_SPECIAL][20] = cpu.r4300i_SPECIAL_DSLLV;
        ops[R4300I_SPECIAL][21] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_SPECIAL][22] = cpu.r4300i_SPECIAL_DSRLV;
        ops[R4300I_SPECIAL][23] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_SPECIAL_DSRAV;
        ops[R4300I_SPECIAL][24] = cpu.r4300i_SPECIAL_MULT;
        ops[R4300I_SPECIAL][25] = cpu.r4300i_SPECIAL_MULTU;
        ops[R4300I_SPECIAL][26] = cpu.r4300i_SPECIAL_DIV;
        ops[R4300I_SPECIAL][27] = cpu.r4300i_SPECIAL_DIVU;
        ops[R4300I_SPECIAL][28] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_SPECIAL_DMULT;
        ops[R4300I_SPECIAL][29] = cpu.r4300i_SPECIAL_DMULTU;
        ops[R4300I_SPECIAL][30] = cpu.r4300i_SPECIAL_DDIV;
        ops[R4300I_SPECIAL][31] = cpu.r4300i_SPECIAL_DDIVU;
        ops[R4300I_SPECIAL][32] = cpu.r4300i_SPECIAL_ADD;
        ops[R4300I_SPECIAL][33] = cpu.r4300i_SPECIAL_ADDU;
        ops[R4300I_SPECIAL][34] = cpu.r4300i_SPECIAL_SUB;
        ops[R4300I_SPECIAL][35] = cpu.r4300i_SPECIAL_SUBU;
        ops[R4300I_SPECIAL][36] = cpu.r4300i_SPECIAL_AND;
        ops[R4300I_SPECIAL][37] = cpu.r4300i_SPECIAL_OR;
        ops[R4300I_SPECIAL][38] = cpu.r4300i_SPECIAL_XOR;
        ops[R4300I_SPECIAL][39] = cpu.r4300i_SPECIAL_NOR;
        ops[R4300I_SPECIAL][40] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_SPECIAL][41] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_SPECIAL][42] = cpu.r4300i_SPECIAL_SLT;
        ops[R4300I_SPECIAL][43] = cpu.r4300i_SPECIAL_SLTU;
        ops[R4300I_SPECIAL][44] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_SPECIAL_DADD;
        ops[R4300I_SPECIAL][45] = cpu.r4300i_SPECIAL_DADDU;
        ops[R4300I_SPECIAL][46] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_SPECIAL_DSUB;
        ops[R4300I_SPECIAL][47] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_SPECIAL_DSUBU;
        ops[R4300I_SPECIAL][48] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_SPECIAL][49] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_SPECIAL][50] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_SPECIAL][51] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_SPECIAL][52] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_SPECIAL_TEQ;
        ops[R4300I_SPECIAL][53] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_SPECIAL][54] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_SPECIAL][55] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_SPECIAL][56] = cpu.r4300i_SPECIAL_DSLL;
        ops[R4300I_SPECIAL][57] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_SPECIAL][58] = cpu.r4300i_SPECIAL_DSRL;
        ops[R4300I_SPECIAL][59] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_SPECIAL_DSRA;
        ops[R4300I_SPECIAL][60] = cpu.r4300i_SPECIAL_DSLL32;
        ops[R4300I_SPECIAL][61] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_SPECIAL][62] = cpu.r4300i_SPECIAL_DSRL32;
        ops[R4300I_SPECIAL][63] = cpu.r4300i_SPECIAL_DSRA32;
        
        ops[R4300I_REGIMM][ 0] = cpu.r4300i_REGIMM_BLTZ;
        ops[R4300I_REGIMM][ 1] = cpu.r4300i_REGIMM_BGEZ;
        ops[R4300I_REGIMM][ 2] = cpu.r4300i_REGIMM_BLTZL;
        ops[R4300I_REGIMM][ 3] = cpu.r4300i_REGIMM_BGEZL;
        ops[R4300I_REGIMM][ 4] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][ 5] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][ 6] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][ 7] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][ 8] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][ 9] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][10] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][11] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][12] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][13] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][14] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][15] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][16] = cpu.r4300i_REGIMM_BLTZAL;
        ops[R4300I_REGIMM][17] = cpu.r4300i_REGIMM_BGEZAL;
        ops[R4300I_REGIMM][18] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][19] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][20] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][21] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][22] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][23] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][24] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][25] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][26] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][27] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][28] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][29] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][30] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_REGIMM][31] = cpu.R4300i_UnknownOpcode;
        
        ops[R4300I_COP0][ 0] = cpu.r4300i_COP0_MF;
        ops[R4300I_COP0][ 1] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0][ 2] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0][ 3] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0][ 4] = cpu.r4300i_COP0_MT;
        ops[R4300I_COP0][ 5] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0][ 6] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0][ 7] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0][ 8] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0][ 9] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0][10] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0][11] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0][12] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0][13] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0][14] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0][15] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0][16] = cpu.R4300i_opcode_COP0_CO;
        ops[R4300I_COP0][17] = cpu.R4300i_opcode_COP0_CO;
        ops[R4300I_COP0][18] = cpu.R4300i_opcode_COP0_CO;
        ops[R4300I_COP0][19] = cpu.R4300i_opcode_COP0_CO;
        ops[R4300I_COP0][20] = cpu.R4300i_opcode_COP0_CO;
        ops[R4300I_COP0][21] = cpu.R4300i_opcode_COP0_CO;
        ops[R4300I_COP0][22] = cpu.R4300i_opcode_COP0_CO;
        ops[R4300I_COP0][23] = cpu.R4300i_opcode_COP0_CO;
        ops[R4300I_COP0][24] = cpu.R4300i_opcode_COP0_CO;
        ops[R4300I_COP0][25] = cpu.R4300i_opcode_COP0_CO;
        ops[R4300I_COP0][26] = cpu.R4300i_opcode_COP0_CO;
        ops[R4300I_COP0][27] = cpu.R4300i_opcode_COP0_CO;
        ops[R4300I_COP0][28] = cpu.R4300i_opcode_COP0_CO;
        ops[R4300I_COP0][29] = cpu.R4300i_opcode_COP0_CO;
        ops[R4300I_COP0][30] = cpu.R4300i_opcode_COP0_CO;
        ops[R4300I_COP0][31] = cpu.R4300i_opcode_COP0_CO;
        
        ops[R4300I_COP0_FUNCTION][ 0] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][ 1] = cop0.r4300i_COP0_CO_TLBR;
        ops[R4300I_COP0_FUNCTION][ 2] = cop0.r4300i_COP0_CO_TLBWI;
        ops[R4300I_COP0_FUNCTION][ 3] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][ 4] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][ 5] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][ 6] = cop0.r4300i_COP0_CO_TLBWR;
        ops[R4300I_COP0_FUNCTION][ 7] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][ 8] = cop0.r4300i_COP0_CO_TLBP;
        ops[R4300I_COP0_FUNCTION][ 9] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][10] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][11] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][12] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][13] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][14] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][15] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][16] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][17] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][18] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][19] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][20] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][21] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][22] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][23] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][24] = cpu.r4300i_COP0_CO_ERET;
        ops[R4300I_COP0_FUNCTION][25] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][26] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][27] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][28] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][29] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][30] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][31] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][32] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][33] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][34] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][35] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][36] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][37] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][38] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][39] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][40] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][41] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][42] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][43] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][44] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][45] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][46] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][47] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][48] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][49] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][50] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][51] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][52] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][53] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][54] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][55] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][56] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][57] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][58] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][59] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][60] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][61] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][62] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP0_FUNCTION][63] = cpu.R4300i_UnknownOpcode;
        
        ops[R4300I_COP1][ 0] = cpu.r4300i_COP1_MF;
        ops[R4300I_COP1][ 1] = cpu.r4300i_COP1_DMF;
        ops[R4300I_COP1][ 2] = cpu.r4300i_COP1_CF;
        ops[R4300I_COP1][ 3] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1][ 4] = cpu.r4300i_COP1_MT;
        ops[R4300I_COP1][ 5] = cpu.r4300i_COP1_DMT;
        ops[R4300I_COP1][ 6] = cpu.r4300i_COP1_CT;
        ops[R4300I_COP1][ 7] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1][ 8] = cpu.R4300i_opcode_COP1_BC;
        ops[R4300I_COP1][ 9] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1][10] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1][11] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1][12] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1][13] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1][14] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1][15] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1][16] = cpu.R4300i_opcode_COP1_S;
        ops[R4300I_COP1][17] = cpu.R4300i_opcode_COP1_D;
        ops[R4300I_COP1][18] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1][19] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1][20] = cpu.R4300i_opcode_COP1_W;
        ops[R4300I_COP1][21] = cpu.R4300i_opcode_COP1_L;
        ops[R4300I_COP1][22] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1][23] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1][24] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1][25] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1][26] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1][27] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1][28] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1][29] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1][30] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1][31] = cpu.R4300i_UnknownOpcode;
        
        ops[R4300I_COP1_BC][ 0] = cpu.r4300i_COP1_BCF;
        ops[R4300I_COP1_BC][ 1] = cpu.r4300i_COP1_BCT;
        ops[R4300I_COP1_BC][ 2] = cpu.r4300i_COP1_BCFL;
        ops[R4300I_COP1_BC][ 3] = cpu.r4300i_COP1_BCTL;
        ops[R4300I_COP1_BC][ 4] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][ 5] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][ 6] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][ 7] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][ 8] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][ 9] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][10] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][11] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][12] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][13] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][14] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][15] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][16] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][17] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][18] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][19] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][20] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][21] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][22] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][23] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][24] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][25] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][26] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][27] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][28] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][29] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][30] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_BC][31] = cpu.R4300i_UnknownOpcode;
        
        ops[R4300I_COP1_S][ 0] = cop1.r4300i_COP1_S_ADD;
        ops[R4300I_COP1_S][ 1] = cop1.r4300i_COP1_S_SUB;
        ops[R4300I_COP1_S][ 2] = cop1.r4300i_COP1_S_MUL;
        ops[R4300I_COP1_S][ 3] = cop1.r4300i_COP1_S_DIV;
        ops[R4300I_COP1_S][ 4] = cop1.r4300i_COP1_S_SQRT;
        ops[R4300I_COP1_S][ 5] = cop1.r4300i_COP1_S_ABS;
        ops[R4300I_COP1_S][ 6] = cop1.r4300i_COP1_S_MOV;
        ops[R4300I_COP1_S][ 7] = cop1.r4300i_COP1_S_NEG;
        ops[R4300I_COP1_S][ 8] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][ 9] = cop1.r4300i_COP1_S_TRUNC_L;
        ops[R4300I_COP1_S][10] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][11] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][12] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_S_ROUND_W;
        ops[R4300I_COP1_S][13] = cop1.r4300i_COP1_S_TRUNC_W;
        ops[R4300I_COP1_S][14] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][15] = cop1.r4300i_COP1_S_FLOOR_W;
        ops[R4300I_COP1_S][16] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][17] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][18] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][19] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][20] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][21] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][22] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][23] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][24] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][25] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][26] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][27] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][28] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][29] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][30] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][31] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][32] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][33] = cop1.r4300i_COP1_S_CVT_D;
        ops[R4300I_COP1_S][34] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][35] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][36] = cop1.r4300i_COP1_S_CVT_W;
        ops[R4300I_COP1_S][37] = cop1.r4300i_COP1_S_CVT_L;
        ops[R4300I_COP1_S][38] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][39] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][40] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][41] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][42] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][43] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][44] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][45] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][46] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][47] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_S][48] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_S_CMP;
        ops[R4300I_COP1_S][49] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_S_CMP;
        ops[R4300I_COP1_S][50] = cop1.r4300i_COP1_S_CMP;
        ops[R4300I_COP1_S][51] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_S_CMP;
        ops[R4300I_COP1_S][52] = cop1.r4300i_COP1_S_CMP;
        ops[R4300I_COP1_S][53] = cop1.r4300i_COP1_S_CMP;
        ops[R4300I_COP1_S][54] = cop1.r4300i_COP1_S_CMP;
        ops[R4300I_COP1_S][55] = cop1.r4300i_COP1_S_CMP;
        ops[R4300I_COP1_S][56] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_S_CMP;
        ops[R4300I_COP1_S][57] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_S_CMP;
        ops[R4300I_COP1_S][58] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_S_CMP;
        ops[R4300I_COP1_S][59] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_S_CMP;
        ops[R4300I_COP1_S][60] = cop1.r4300i_COP1_S_CMP;
        ops[R4300I_COP1_S][61] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_S_CMP;
        ops[R4300I_COP1_S][62] = cop1.r4300i_COP1_S_CMP;
        ops[R4300I_COP1_S][63] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_S_CMP;
        
        ops[R4300I_COP1_D][ 0] = cop1.r4300i_COP1_D_ADD;
        ops[R4300I_COP1_D][ 1] = cop1.r4300i_COP1_D_SUB;
        ops[R4300I_COP1_D][ 2] = cop1.r4300i_COP1_D_MUL;
        ops[R4300I_COP1_D][ 3] = cop1.r4300i_COP1_D_DIV;
        ops[R4300I_COP1_D][ 4] = cop1.r4300i_COP1_D_SQRT;
        ops[R4300I_COP1_D][ 5] = cop1.r4300i_COP1_D_ABS;
        ops[R4300I_COP1_D][ 6] = cop1.r4300i_COP1_D_MOV;
        ops[R4300I_COP1_D][ 7] = cop1.r4300i_COP1_D_NEG;
        ops[R4300I_COP1_D][ 8] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][ 9] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][10] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][11] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][12] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_D_ROUND_W;
        ops[R4300I_COP1_D][13] = cop1.r4300i_COP1_D_TRUNC_W;
        ops[R4300I_COP1_D][14] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][15] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][16] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][17] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][18] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][19] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][20] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][21] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][22] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][23] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][24] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][25] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][26] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][27] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][28] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][29] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][30] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][31] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][32] = cop1.r4300i_COP1_D_CVT_S;
        ops[R4300I_COP1_D][33] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][34] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][35] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][36] = cop1.r4300i_COP1_D_CVT_W;
        ops[R4300I_COP1_D][37] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_D_CVT_L;
        ops[R4300I_COP1_D][38] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][39] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][40] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][41] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][42] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][43] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][44] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][45] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][46] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][47] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_D][48] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_D_CMP;
        ops[R4300I_COP1_D][49] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_D_CMP;
        ops[R4300I_COP1_D][50] = cop1.r4300i_COP1_D_CMP;
        ops[R4300I_COP1_D][51] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_D_CMP;
        ops[R4300I_COP1_D][52] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_D_CMP;
        ops[R4300I_COP1_D][53] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_D_CMP;
        ops[R4300I_COP1_D][54] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_D_CMP;
        ops[R4300I_COP1_D][55] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_D_CMP;
        ops[R4300I_COP1_D][56] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_D_CMP;
        ops[R4300I_COP1_D][57] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_D_CMP;
        ops[R4300I_COP1_D][58] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_D_CMP;
        ops[R4300I_COP1_D][59] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_D_CMP;
        ops[R4300I_COP1_D][60] = cop1.r4300i_COP1_D_CMP;
        ops[R4300I_COP1_D][61] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_D_CMP;
        ops[R4300I_COP1_D][62] = cop1.r4300i_COP1_D_CMP;
        ops[R4300I_COP1_D][63] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_D_CMP;
        
        ops[R4300I_COP1_W][ 0] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][ 1] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][ 2] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][ 3] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][ 4] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][ 5] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][ 6] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][ 7] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][ 8] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][ 9] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][10] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][11] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][12] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][13] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][14] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][15] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][16] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][17] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][18] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][19] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][20] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][21] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][22] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][23] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][24] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][25] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][26] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][27] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][28] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][29] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][30] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][31] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][32] = cop1.r4300i_COP1_W_CVT_S;
        ops[R4300I_COP1_W][33] = cop1.r4300i_COP1_W_CVT_D;
        ops[R4300I_COP1_W][34] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][35] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][36] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][37] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][38] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][39] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][40] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][41] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][42] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][43] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][44] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][45] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][46] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][47] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][48] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][49] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][50] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][51] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][52] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][53] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][54] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][55] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][56] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][57] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][58] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][59] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][60] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][61] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][62] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_W][63] = cpu.R4300i_UnknownOpcode;
        
        ops[R4300I_COP1_L][ 0] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][ 1] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][ 2] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][ 3] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][ 4] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][ 5] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][ 6] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][ 7] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][ 8] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][ 9] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][10] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][11] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][12] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][13] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][14] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][15] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][16] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][17] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][18] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][19] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][20] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][21] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][22] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][23] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][24] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][25] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][26] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][27] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][28] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][29] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][30] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][31] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][32] = cop1.r4300i_COP1_L_CVT_S;
        ops[R4300I_COP1_L][33] = cpu.R4300i_UnknownOpcode; //cpu.r4300i_COP1_L_CVT_D;
        ops[R4300I_COP1_L][34] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][35] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][36] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][37] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][38] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][39] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][40] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][41] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][42] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][43] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][44] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][45] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][46] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][47] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][48] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][49] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][50] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][51] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][52] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][53] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][54] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][55] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][56] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][57] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][58] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][59] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][60] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][61] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][62] = cpu.R4300i_UnknownOpcode;
        ops[R4300I_COP1_L][63] = cpu.R4300i_UnknownOpcode;
        
        return ops;
    }
    
    public static void main(String[] args) {
        try {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        } catch(Exception ex) { ex.printStackTrace(); }
        
        Debug.init();
        
        int WindowWidth = 640;
        int WindowHeight = 480;
        
        // Registers
        regs = new Registers();
        
//        // Cpu
//        Cpu cpu = new Cpu(regs);
//        //DebugCpu cpu = new DebugCpu();
        
        // Memory
        try {
            mem = new Memory(0x00800000); // (8MB)
        } catch (MemoryException ex) {
            System.err.printf("Failed to allocate Memory\n");
            exit("Failed to initialize application");
        }
        
        // Cpu
        cpu = new Cpu(regs, mem);
        //DebugCpu cpu = new DebugCpu(regs, mem);
        
//        // Registers
//        Registers regs = new Registers();
        
        // Rsp
        rsp = new Rsp(cpu.CheckInterrupts, regs, mem.RDRAM, mem.DMEM, mem.IMEM);
        
        // Video
        video = new Video(cpu.CheckInterrupts, regs, mem.RDRAM, mem.DMEM, mem.IMEM);
        
        // Audio
        audio = new Audio(cpu.AiCheckInterrupts, mem.RDRAM, mem.DMEM, mem.IMEM);
        
        // Pif
        pif = new Pif(cpu.CheckInterrupts, regs);
        
        // Dma
        dma = new Dma(cpu.CheckInterrupts, regs, mem.RDRAM, mem.DMEM);
        
//        ShowUnhandledMemory = Default_ShowUnhandledMemory;
//        ShowTLBMisses = Default_ShowTLBMisses;
//        ShowPifRamErrors = Default_ShowPifRamErrors;
        
        hMainWindow = new JFrame("N64");
        hMainWindow.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                exit("Window Closed.");
            }
        });
        hMainWindow.addKeyListener(new KeyListener() {
            public void keyReleased(KeyEvent e) {
                if (pif.controllerPlugin != null)
                    pif.controllerPlugin.wmKeyUp(e.getKeyCode(), 0);
            }
            
            public void keyPressed(KeyEvent e) {
                if (pif.controllerPlugin != null)
                    pif.controllerPlugin.wmKeyDown(e.getKeyCode(), 0);
            }
            
            public void keyTyped(KeyEvent e) {
            }
        });
        
        hMainWindow.setJMenuBar(new MenuBar());
        
        panel = new JPanel();
        panel.setPreferredSize(new java.awt.Dimension(WindowWidth, WindowHeight));
        hMainWindow.getContentPane().add(panel);
        
//        // Choose a rom
//        File romFile = null;
//        JFileChooser fileChooser = new JFileChooser();
//        fileChooser.setDialogTitle("Open ROM");
//        int returnVal = fileChooser.showOpenDialog(null);
//        if (returnVal == JFileChooser.APPROVE_OPTION)
//            romFile = fileChooser.getSelectedFile();
//        else
//            exit("No file chosen.");
        
        hMainWindow.pack();
        hMainWindow.setVisible(true);
        
        // Plugins
        URL url = null;
        File file = new File("./plugins.properties");
        if (file.exists()) {
            try {
                url = file.toURI().toURL();
            } catch(Exception ex) { }
        }
        else {
            url = ResourceAnchor.class.getClassLoader().getResource("resources/plugins.properties");
        }

        if (url == null) {
            System.out.println("can't find resource: " + "plugins.properties");
            return;
        }
        
        Properties cfg = new Properties();
        try {
//            cfg.load(new FileInputStream(file));
            cfg.load(url.openStream());
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        video.setupPlugin(hMainWindow, cfg);
        audio.setupPlugin(hMainWindow, cfg);
//        rsp.setupPlugin(
//                video.gfxPlugin.getProcessDList(),
//                audio.audioPlugin.getProcessAList(),
//                video.gfxPlugin.getProcessRDPList());
//        rsp.setupPlugin(video.gfxPlugin, audio.audioPlugin);
        rsp.setupPlugin(hMainWindow, cfg, video.gfxPlugin, audio.audioPlugin);
        pif.setupPlugin(hMainWindow, cfg);
        if (DEBUG) System.out.printf("Plugins Setup\n");
        
        //openRom();
//        // Choose a rom
//        File romFile = null;
//        JFileChooser fileChooser = new JFileChooser();
//        fileChooser.setDialogTitle("Open ROM");
//        int returnVal = fileChooser.showOpenDialog(null);
//        if (returnVal == JFileChooser.APPROVE_OPTION)
//            romFile = fileChooser.getSelectedFile();
//        else
//            exit("No file chosen.");
//        
//        CloseCpu();
//        
//        Rom rom = new Rom(romFile);
//        if (DEBUG) System.out.printf("Opened rom: %s\n", romFile);
//        
//        // Save Directory
//        File Directory = new File(AUTO_SAVE_DIR);
//        if (!Directory.exists())
//            Directory.mkdirs();
//        String romName = romFile.getName().substring(0, romFile.getName().lastIndexOf('.'));
//        String path = Directory.getAbsolutePath() + "/" + romName;
//        
//        // Save Chips
//        sram = new Sram(path + ".sra");
//        mempak = new Mempak(path + ".mpk");
//        flashRam = new FlashRam(path + ".fla");
//        eeprom = new Eeprom(path + ".eep");
//        
//        pif.setSaveChips(eeprom, mempak);
//        pif.showPifRamErrors(ShowPifRamErrors);
//        dma.setSaveChips(sram, flashRam);
//        dma.showUnhandledMemory(ShowUnhandledMemory);
//        
//        dma.connect(pif, rsp, rom.data);
//        mem.map(cpu.CheckInterrupts, rsp, video, audio, pif, dma, flashRam, regs, rom.data);
//        
//        // Could be made ROM dependent
//        int countPerOp = -1;
//        if (countPerOp < 1)  { countPerOp = Default_CountPerOp; }
//        if (countPerOp > 6)  { countPerOp = Default_CountPerOp; }
//        regs.saveUsing = Registers.AUTO;
//        UseTlb = Default_UseTlb;
//        CacheInstructions = Default_CacheInstructions;
//        
//        // Cop0
//        Cop0 cop0 = null;
//        try {
//            cop0 = new Cop0(countPerOp);
//        } catch (TlbException ex) {
//            System.err.printf("Failed to allocate Memory\n");
//            exit("Failed to initialize application");
//        }
//        cop0.useTlb(UseTlb);
//        cop0.setTimerInterrupts(cpu.CheckInterrupts, pif.timerInterrupt, dma.timerInterrupt, video.timerInterrupt);
//        video.setTimer(cop0);
//        
//        // Cop1
//        Cop1 cop1 = new Cop1();
//        
//        cpu.cacheInstructions(CacheInstructions);
//        cpu.connect(cop0, cop1);
//        cpu.setOps(buildCpuOps(cpu, cop0, cop1));
//        //cpu.setOps(DebugCpu.buildDebugCpuOps(cpu));
//        
//        CloseCpu();
//        
//        initialize(pif, rom, mem, cpu);
//        
//        cpuStartTime = System.currentTimeMillis();
//        if (video.gfxPlugin != null)
//            video.gfxPlugin.romOpen();
//        if (pif.controllerPlugin != null)
//            pif.controllerPlugin.romOpen();
//        
//        if (DEBUG) System.out.printf("Starting main emulation loop ...\n");
//        cpu.startEmulation();
    }
    
    public static void openRom() {
//        ShowUnhandledMemory = Default_ShowUnhandledMemory;
//        ShowTLBMisses = Default_ShowTLBMisses;
//        ShowPifRamErrors = Default_ShowPifRamErrors;
        
        // Choose a rom
        File romFile = null;
        JFileChooser fileChooser = new JFileChooser();
        fileChooser.setDialogTitle("Open ROM");
        int returnVal = fileChooser.showOpenDialog(null);
        if (returnVal == JFileChooser.APPROVE_OPTION)
            romFile = fileChooser.getSelectedFile();
        else
            //exit("No file chosen.");
            return;
        
        CloseCpu();
        
        Rom rom = new Rom(romFile);
        if (DEBUG) System.out.printf("Opened rom: %s\n", romFile);
        
        // Save Directory
        File Directory = new File(AUTO_SAVE_DIR);
        if (!Directory.exists())
            Directory.mkdirs();
        String romName = romFile.getName().substring(0, romFile.getName().lastIndexOf('.'));
        String path = Directory.getAbsolutePath() + "/" + romName;
        
        // Save Chips
        sram = new Sram(path + ".sra");
        mempak = new Mempak(path + ".mpk");
        flashRam = new FlashRam(path + ".fla");
        eeprom = new Eeprom(path + ".eep");
        
        pif.setSaveChips(eeprom, mempak);
        pif.showPifRamErrors(ShowPifRamErrors);
        dma.setSaveChips(sram, flashRam);
        dma.showUnhandledMemory(ShowUnhandledMemory);
        
        dma.connect(pif, rsp, rom.data);
        mem.map(cpu.CheckInterrupts, rsp, video, audio, pif, dma, flashRam, regs, rom.data);
        
        // Could be made ROM dependent
        int countPerOp = -1;
        if (countPerOp < 1)  { countPerOp = Default_CountPerOp; }
        if (countPerOp > 6)  { countPerOp = Default_CountPerOp; }
        regs.saveUsing = Registers.AUTO;
        //UseTlb = Default_UseTlb;
        //CacheInstructions = Default_CacheInstructions;
        
        // Cop0
        Cop0 cop0 = null;
        try {
            cop0 = new Cop0(countPerOp);
        } catch (TlbException ex) {
            System.err.printf("Failed to allocate Memory\n");
            exit("Failed to initialize application");
        }
        cop0.useTlb(UseTlb);
        cop0.setTimerInterrupts(cpu.CheckInterrupts, pif.timerInterrupt, dma.timerInterrupt, video.timerInterrupt);
        video.setTimer(cop0);
        video.setFrameLimit(FrameLimit);
        mem.setTimer(cop0);
        
        // Cop1
        Cop1 cop1 = new Cop1();
        
        cpu.cacheInstructions(CacheInstructions);
        cpu.connect(cop0, cop1);
        cpu.setOps(buildCpuOps(cpu, cop0, cop1));
        //cpu.setOps(DebugCpu.buildDebugCpuOps(cpu));
        
        CloseCpu();
        
        initialize(pif, rom, mem, cpu);
        
        cpuStartTime = System.currentTimeMillis();
        if (video.gfxPlugin != null)
            video.gfxPlugin.romOpen();
        if (pif.controllerPlugin != null)
            pif.controllerPlugin.romOpen();
        
        if (!UseAudio && audio.audioPlugin != null) {
            audio.audioPlugin.closePlugin();
            audio.audioPlugin = null;
//            rsp.setHLE(video.gfxPlugin, audio.audioPlugin);
        }
        
        if (DEBUG) System.out.printf("Starting main emulation loop ...\n");
        cpu.startEmulation();
    }
    
    private static void initialize(Pif pif, Rom rom, Memory mem, Cpu cpu) {
        boolean usePif = pif.loadPifRom(rom.data.get(0x3E));
        int country = rom.data.get(0x3E);
        int cicChip = pif.getCicChipID(rom.data);
        if (cicChip < 0) {
            System.err.printf("Unknown Cic Chip: %d", cicChip);
            cicChip = 2;
        }

        if (usePif) {
            System.out.println("Using PIF");
            cpu.pc = 0xBFC00000;
            pif.initalizeRegisters(cicChip);
        } else {
            if (rom.data != null) {
                rom.data.position(0x040);
                mem.DMEM.position(0x040);
                for (int i=0; i<0xFBC; i++)
                    mem.DMEM.put(rom.data.get());
            }
            cpu.pc = 0xA4000040;
            
//            cpu.GPR[0] = 0x0000000000000000L;
//            cpu.GPR[6] = 0xFFFFFFFFA4001F0CL;
//            cpu.GPR[7] = 0xFFFFFFFFA4001F08L;
//            cpu.GPR[8] = 0x00000000000000C0L;
//            cpu.GPR[9] = 0x0000000000000000L;
//            cpu.GPR[10] = 0x0000000000000040L;
//            cpu.GPR[11] = 0xFFFFFFFFA4000040L;
//            cpu.GPR[16] = 0x0000000000000000L;
//            cpu.GPR[17] = 0x0000000000000000L;
//            cpu.GPR[18] = 0x0000000000000000L;
//            cpu.GPR[19] = 0x0000000000000000L;
//            cpu.GPR[21] = 0x0000000000000000L;
//            cpu.GPR[26] = 0x0000000000000000L;
//            cpu.GPR[27] = 0x0000000000000000L;
//            cpu.GPR[28] = 0x0000000000000000L;
//            cpu.GPR[29] = 0xFFFFFFFFA4001FF0L;
//            cpu.GPR[30] = 0x0000000000000000L;
            cpu.GPR[0] = 0x00000000;
            cpu.GPR[6] = 0xA4001F0C;
            cpu.GPR[7] = 0xA4001F08;
            cpu.GPR[8] = 0x000000C0;
            cpu.GPR[9] = 0x00000000;
            cpu.GPR[10] = 0x00000040;
            cpu.GPR[11] = 0xA4000040;
            cpu.GPR[16] = 0x00000000;
            cpu.GPR[17] = 0x00000000;
            cpu.GPR[18] = 0x00000000;
            cpu.GPR[19] = 0x00000000;
            cpu.GPR[21] = 0x00000000;
            cpu.GPR[26] = 0x00000000;
            cpu.GPR[27] = 0x00000000;
            cpu.GPR[28] = 0x00000000;
            cpu.GPR[29] = 0xA4001FF0;
            cpu.GPR[30] = 0x00000000;
            
            switch (country) {
                case 0x44: //Germany
                case 0x46: //french
                case 0x49: //Italian
                case 0x50: //Europe
                case 0x53: //Spanish
                case 0x55: //Australia
                case 0x58: // ????
                case 0x59: // X (PAL)
                    switch (cicChip) {
                        case 2:
//                            cpu.GPR[5] = 0xFFFFFFFFC0F1D859L;
//                            cpu.GPR[14] = 0x000000002DE108EAL;
//                            cpu.GPR[24] = 0x0000000000000000L;
                            cpu.GPR[5] = 0xC0F1D859;
                            cpu.GPR[14] = 0x2DE108EA;
                            cpu.GPR[24] = 0x00000000;
                            break;
                        case 3:
//                            cpu.GPR[5] = 0xFFFFFFFFD4646273L;
//                            cpu.GPR[14] = 0x000000001AF99984L;
//                            cpu.GPR[24] = 0x0000000000000000L;
                            cpu.GPR[5] = 0xD4646273;
                            cpu.GPR[14] = 0x1AF99984;
                            cpu.GPR[24] = 0x00000000;
                            break;
                        case 5:
                            mem.DMEM.putInt(0x1004, 0xBDA807FC);
//                            cpu.GPR[5] = 0xFFFFFFFFDECAAAD1L;
//                            cpu.GPR[14] = 0x000000000CF85C13L;
//                            cpu.GPR[24] = 0x0000000000000002L;
                            cpu.GPR[5] = 0xDECAAAD1;
                            cpu.GPR[14] = 0x0CF85C13;
                            cpu.GPR[24] = 0x00000002;
                            break;
                        case 6:
//                            cpu.GPR[5] = 0xFFFFFFFFB04DC903L;
//                            cpu.GPR[14] = 0x000000001AF99984L;
//                            cpu.GPR[24] = 0x0000000000000002L;
                            cpu.GPR[5] = 0xB04DC903;
                            cpu.GPR[14] = 0x1AF99984;
                            cpu.GPR[24] = 0x00000002;
                            break;
                    }
                    
//                    cpu.GPR[20] = 0x0000000000000000L;
//                    cpu.GPR[23] = 0x0000000000000006L;
//                    cpu.GPR[31] = 0xFFFFFFFFA4001554L;
                    cpu.GPR[20] = 0x00000000;
                    cpu.GPR[23] = 0x00000006;
                    cpu.GPR[31] = 0xA4001554;
                    break;
                case 0x37: // 7 (Beta)
                case 0x41: // ????
                case 0x45: //USA
                case 0x4A: //Japan
                default:
                    switch (cicChip) {
                        case 2:
//                            cpu.GPR[5] = 0xFFFFFFFFC95973D5L;
//                            cpu.GPR[14] = 0x000000002449A366L;
                            cpu.GPR[5] = 0xC95973D5;
                            cpu.GPR[14] = 0x2449A366;
                            break;
                        case 3:
//                            cpu.GPR[5] = 0xFFFFFFFF95315A28L;
//                            cpu.GPR[14] = 0x000000005BACA1DFL;
                            cpu.GPR[5] = 0x95315A28;
                            cpu.GPR[14] = 0x5BACA1DF;
                            break;
                        case 5:
                            mem.DMEM.putInt(0x1004, 0x8DA807FC);
//                            cpu.GPR[5] = 0x000000005493FB9AL;
//                            cpu.GPR[14] = 0xFFFFFFFFC2C20384L;
                            cpu.GPR[5] = 0x5493FB9A;
                            cpu.GPR[14] = 0xC2C20384;
                        case 6:
//                            cpu.GPR[5] = 0xFFFFFFFFE067221FL;
//                            cpu.GPR[14] = 0x000000005CD2B70FL;
                            cpu.GPR[5] = 0xE067221F;
                            cpu.GPR[14] = 0x5CD2B70F;
                            break;
                    }
//                    cpu.GPR[20] = 0x0000000000000001L;
//                    cpu.GPR[23] = 0x0000000000000000L;
//                    cpu.GPR[24] = 0x0000000000000003L;
//                    cpu.GPR[31] = 0xFFFFFFFFA4001550L;
                    cpu.GPR[20] = 0x00000001;
                    cpu.GPR[23] = 0x00000000;
                    cpu.GPR[24] = 0x00000003;
                    cpu.GPR[31] = 0xA4001550;
            }
            
            switch (cicChip) {
                case 1:
//                    cpu.GPR[22] = 0x000000000000003FL;
                    cpu.GPR[22] = 0x0000003F;
                    break;
                case 2:
//                    cpu.GPR[1] = 0x0000000000000001L;
//                    cpu.GPR[2] = 0x000000000EBDA536L;
//                    cpu.GPR[3] = 0x000000000EBDA536L;
//                    cpu.GPR[4] = 0x000000000000A536L;
//                    cpu.GPR[12] = 0xFFFFFFFFED10D0B3L;
//                    cpu.GPR[13] = 0x000000001402A4CCL;
//                    cpu.GPR[15] = 0x000000003103E121L;
//                    cpu.GPR[22] = 0x000000000000003FL;
//                    cpu.GPR[25] = 0xFFFFFFFF9DEBB54FL;
                    cpu.GPR[1] = 0x00000001;
                    cpu.GPR[2] = 0x0EBDA536;
                    cpu.GPR[3] = 0x0EBDA536;
                    cpu.GPR[4] = 0x0000A536;
                    cpu.GPR[12] = 0xED10D0B3;
                    cpu.GPR[13] = 0x1402A4CC;
                    cpu.GPR[15] = 0x3103E121;
                    cpu.GPR[22] = 0x0000003F;
                    cpu.GPR[25] = 0x9DEBB54F;
                    break;
                case 3:
//                    cpu.GPR[1] = 0x0000000000000001L;
//                    cpu.GPR[2] = 0x0000000049A5EE96L;
//                    cpu.GPR[3] = 0x0000000049A5EE96L;
//                    cpu.GPR[4] = 0x000000000000EE96L;
//                    cpu.GPR[12] = 0xFFFFFFFFCE9DFBF7L;
//                    cpu.GPR[13] = 0xFFFFFFFFCE9DFBF7L;
//                    cpu.GPR[15] = 0x0000000018B63D28L;
//                    cpu.GPR[22] = 0x0000000000000078L;
//                    cpu.GPR[25] = 0xFFFFFFFF825B21C9L;
                    cpu.GPR[1] = 0x00000001;
                    cpu.GPR[2] = 0x49A5EE96;
                    cpu.GPR[3] = 0x49A5EE96;
                    cpu.GPR[4] = 0x0000EE96;
                    cpu.GPR[12] = 0xCE9DFBF7;
                    cpu.GPR[13] = 0xCE9DFBF7;
                    cpu.GPR[15] = 0x18B63D28;
                    cpu.GPR[22] = 0x00000078;
                    cpu.GPR[25] = 0x825B21C9;
                    break;
                case 5:
                    mem.DMEM.putInt(0x1000, 0x3C0DBFC0);
                    mem.DMEM.putInt(0x1008, 0x25AD07C0);
                    mem.DMEM.putInt(0x100C, 0x31080080);
                    mem.DMEM.putInt(0x1010, 0x5500FFFC);
                    mem.DMEM.putInt(0x1014, 0x3C0DBFC0);
                    mem.DMEM.putInt(0x1018, 0x8DA80024);
                    mem.DMEM.putInt(0x101C, 0x3C0BB000);
                    
//                    cpu.GPR[1] = 0x0000000000000000L;
//                    cpu.GPR[2] = 0xFFFFFFFFF58B0FBFL;
//                    cpu.GPR[3] = 0xFFFFFFFFF58B0FBFL;
//                    cpu.GPR[4] = 0x0000000000000FBFL;
//                    cpu.GPR[12] = 0xFFFFFFFF9651F81EL;
//                    cpu.GPR[13] = 0x000000002D42AAC5L;
//                    cpu.GPR[15] = 0x0000000056584D60L;
//                    cpu.GPR[22] = 0x0000000000000091L;
//                    cpu.GPR[25] = 0xFFFFFFFFCDCE565FL;
                    cpu.GPR[1] = 0x00000000;
                    cpu.GPR[2] = 0xF58B0FBF;
                    cpu.GPR[3] = 0xF58B0FBF;
                    cpu.GPR[4] = 0x00000FBF;
                    cpu.GPR[12] = 0x9651F81E;
                    cpu.GPR[13] = 0x2D42AAC5;
                    cpu.GPR[15] = 0x56584D60;
                    cpu.GPR[22] = 0x00000091;
                    cpu.GPR[25] = 0xCDCE565F;
                    break;
                case 6:
//                    cpu.GPR[1] = 0x0000000000000000L;
//                    cpu.GPR[2] = 0xFFFFFFFFA95930A4L;
//                    cpu.GPR[3] = 0xFFFFFFFFA95930A4L;
//                    cpu.GPR[4] = 0x00000000000030A4L;
//                    cpu.GPR[12] = 0xFFFFFFFFBCB59510L;
//                    cpu.GPR[13] = 0xFFFFFFFFBCB59510L;
//                    cpu.GPR[15] = 0x000000007A3C07F4L;
//                    cpu.GPR[22] = 0x0000000000000085L;
//                    cpu.GPR[25] = 0x00000000465E3F72L;
                    cpu.GPR[1] = 0x00000000;
                    cpu.GPR[2] = 0xA95930A4;
                    cpu.GPR[3] = 0xA95930A4;
                    cpu.GPR[4] = 0x000030A4;
                    cpu.GPR[12] = 0xBCB59510;
                    cpu.GPR[13] = 0xBCB59510;
                    cpu.GPR[15] = 0x7A3C07F4;
                    cpu.GPR[22] = 0x00000085;
                    cpu.GPR[25] = 0x465E3F72;
                    break;
            }
        }
        if (DEBUG) System.out.printf("R4300i Registers Initialized\n");
    }
    
}
